import ujson
import pickle
from abc import ABCMeta, abstractmethod

__all__ = ['PickleSerializer', 'JsonSerializer', 'AbsSerializer']


class AbsSerializer(object, metaclass=ABCMeta):

    @staticmethod
    @abstractmethod
    def loads(s):
        """Serializer object"""

    @staticmethod
    @abstractmethod
    def dumps(obj):
        """Serializer object"""


class PickleSerializer(AbsSerializer):
    @staticmethod
    def loads(s):
        return pickle.loads(s)

    @staticmethod
    def dumps(obj):
        return pickle.dumps(obj, protocol=-1)


class JsonSerializer(AbsSerializer):
    @staticmethod
    def loads(s):
        return ujson.loads(s)

    @staticmethod
    def dumps(obj):
        return ujson.dumps(obj)




"""
SSL/TLS指纹随机中间件
"""
import random


ORIGIN_CIPHERS = ('ECDH+AESGCM:DH+AESGCM:ECDH+AES256:DH+AES256:ECDH+AES128:DH+AES:ECDH+HIGH:'
                  'DH+HIGH:ECDH+3DES:DH+3DES:RSA+AESGCM:RSA+AES:RSA+HIGH:RSA+3DES:!aNULL:!eNULL:!MD5')


class TLSCiphersMiddleware:
    """SSL/TLS指纹中间件"""

    def __init__(self, ciphers, is_random):
        if ciphers == 'DEFAULT':
            self.ciphers = None

        self.is_random = is_random

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            ciphers=crawler.settings.get('DOWNLOADER_CLIENT_TLS_CIPHERS', 'DEFAULT'),
            is_random=crawler.settings.get('RANDOM_TLS_CIPHERS', False)
        )

    def process_request(self, request, spider):
        if not (self.ciphers or self.is_random):
            return

        ciphers = self.ciphers or ORIGIN_CIPHERS
        if self.is_random:
            ciphers = ciphers.split(":")
            random.shuffle(ciphers)
            ciphers = ":".join(ciphers)
        request.meta['TLS_CIPHERS'] = ciphers



"""
Helper functions for serializing (and deserializing) requests.
"""
from typing import Optional

import aioscrapy
from aioscrapy.utils.request import request_from_dict as _from_dict


def request_to_dict(request: "aioscrapy.Request", spider: Optional["aioscrapy.Spider"] = None) -> dict:
    return request.to_dict(spider=spider)


def request_from_dict(d: dict, spider: Optional["aioscrapy.Spider"] = None) -> "aioscrapy.Request":
    return _from_dict(d, spider=spider)



"""
Base class for Scrapy spiders

See documentation in docs/topics/spiders.rst
"""
import logging
import warnings
from typing import Optional

from aioscrapy import signals
from aioscrapy.http.request import Request
from aioscrapy.http.response import Response
from aioscrapy.utils.url import url_is_from_spider
from aioscrapy.utils.deprecate import method_is_overridden
from aioscrapy.exceptions import DontCloseSpider
from aioscrapy.utils.tools import call_helper


class Spider(object):
    """Base class for scrapy spiders. All spiders must inherit from this
    class.
    """

    name: Optional[str] = None
    proxy: Optional["aioscrapy.proxy.AbsProxy"] = None
    custom_settings: Optional[dict] = None

    def __init__(self, name=None, **kwargs):
        if name is not None:
            self.name = name
        elif not getattr(self, 'name', None):
            raise ValueError(f"{type(self).__name__} must have a name")
        self.__dict__.update(kwargs)
        if not hasattr(self, 'start_urls'):
            self.start_urls = []

    @property
    def logger(self):
        logger = logging.getLogger(self.name)
        return logging.LoggerAdapter(logger, {'spider': self})

    def log(self, message, level=logging.DEBUG, **kw):
        """Log the given message at the given log level

        This helper wraps a log call to the logger within the spider, but you
        can use it directly (e.g. Spider.logger.info('msg')) or use any other
        Python logger too.
        """
        self.logger.log(level, message, **kw)

    @classmethod
    async def from_crawler(cls, crawler, *args, **kwargs):
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        return spider

    def _set_crawler(self, crawler):
        self.crawler = crawler
        self.settings = crawler.settings
        self.close_on_idle = self.settings.get("CLOSE_SPIDER_ON_IDLE", True)
        crawler.signals.connect(self.close, signals.spider_closed)
        crawler.signals.connect(self.spider_idle, signal=signals.spider_idle)

    async def start_requests(self):
        cls = self.__class__
        if not self.start_urls and hasattr(self, 'start_url'):
            raise AttributeError(
                "Crawling could not start: 'start_urls' not found "
                "or empty (but found 'start_url' attribute instead, "
                "did you miss an 's'?)")
        if method_is_overridden(cls, Spider, 'make_requests_from_url'):
            warnings.warn(
                "Spider.make_requests_from_url method is deprecated; it "
                "won't be called in future Scrapy releases. Please "
                "override Spider.start_requests method instead "
                f"(see {cls.__module__}.{cls.__name__}).",
            )
            for url in self.start_urls:
                yield self.make_requests_from_url(url)
        else:
            for url in self.start_urls:
                yield Request(url)

    def make_requests_from_url(self, url):
        """ This method is deprecated. """
        warnings.warn(
            "Spider.make_requests_from_url method is deprecated: "
            "it will be removed and not be called by the default "
            "Spider.start_requests method in future Scrapy releases. "
            "Please override Spider.start_requests method instead."
        )
        return Request(url, dont_filter=True)

    async def _parse(self, response: Response, **kwargs):
        return await call_helper(self.parse, response)

    async def parse(self, response: Response):
        raise NotImplementedError(f'{self.__class__.__name__}.parse callback is not defined')

    @classmethod
    def update_settings(cls, settings):
        settings.setdict(cls.custom_settings or {}, priority='spider')

    @classmethod
    def handles_request(cls, request):
        return url_is_from_spider(request.url, cls)

    @staticmethod
    def close(spider, reason):
        closed = getattr(spider, 'closed', None)
        if callable(closed):
            return closed(reason)

    def __str__(self):
        return f"<{type(self).__name__} {self.name!r} at 0x{id(self):0x}>"

    __repr__ = __str__

    @classmethod
    def start(cls, setting_path=None):
        from aioscrapy.crawler import CrawlerProcess
        from aioscrapy.utils.project import get_project_settings

        settings = get_project_settings()
        if setting_path is not None:
            settings.setmodule(setting_path)
        cp = CrawlerProcess(settings)
        cp.crawl(cls)
        cp.start()

    def spider_idle(self):
        if not self.close_on_idle:
            raise DontCloseSpider



import asyncio
import sys
from typing import Optional, Tuple, List, Union, Type, AnyStr

from aiomultiprocess import Process

from aioscrapy import Spider
from aioscrapy.crawler import CrawlerProcess
from aioscrapy.settings import Settings


def loop_initializer():
    if sys.platform.startswith('win'):
        return asyncio.windows_events.ProactorEventLoop()
    try:
        import uvloop
        return uvloop.EventLoopPolicy()
    except ImportError:
        pass

    return asyncio.new_event_loop()


def multi_process_run(*tasks: Union[Tuple[Type[Spider], Optional[AnyStr]], List]):
    for task in tasks:
        if isinstance(task, list):
            p = Process(target=_single_process_run_async, args=(*task,), loop_initializer=loop_initializer)
        else:
            p = Process(target=_single_process_run_async, args=(task,), loop_initializer=loop_initializer)
        p.start()


async def _single_process_run_async(*tasks: Tuple[Type[Spider], Optional[AnyStr]]):
    cp = CrawlerProcess()
    for spidercls, settings in tasks:
        if isinstance(settings, str):
            instance = Settings()
            instance.setmodule(settings)
            settings = instance
        cp.crawl(spidercls, settings=settings)
    await cp.run()


def single_process_run(*tasks: Tuple[Type[Spider], Optional[AnyStr]]):
    cp = CrawlerProcess()
    for spidercls, settings in tasks:
        if isinstance(settings, str):
            instance = Settings()
            instance.setmodule(settings)
            settings = instance
        cp.crawl(spidercls, settings=settings)
    cp.start()



from typing import Optional, List

import aioscrapy
from aioscrapy.db import db_manager
from aioscrapy.queue import AbsQueue
from aioscrapy.serializer import AbsSerializer
from aioscrapy.utils.misc import load_object


class RabbitMqPriorityQueue(AbsQueue):
    inc_key = 'scheduler/enqueued/rabbitmq'

    @classmethod
    def from_dict(cls, data: dict) -> "RabbitMqPriorityQueue":
        alias: str = data.get("alias", 'queue')
        server: aioscrapy.db.aiorabbitmq.RabbitmqExecutor = db_manager.rabbitmq.executor(alias)
        spider_name: str = data["spider_name"]
        serializer: str = data.get("serializer", "aioscrapy.serializer.JsonSerializer")
        serializer: AbsSerializer = load_object(serializer)
        return cls(
            server,
            key='%(spider)s:requests' % {'spider': spider_name},
            serializer=serializer
        )

    @classmethod
    async def from_spider(cls, spider: aioscrapy.Spider) -> "RabbitMqPriorityQueue":
        alias: str = spider.settings.get("SCHEDULER_QUEUE_ALIAS", 'queue')
        executor: aioscrapy.db.aiorabbitmq.RabbitmqExecutor = db_manager.rabbitmq.executor(alias)
        queue_key: str = spider.settings.get("SCHEDULER_QUEUE_KEY", '%(spider)s:requests')
        serializer: str = spider.settings.get("SCHEDULER_SERIALIZER", "aioscrapy.serializer.JsonSerializer")
        serializer: AbsSerializer = load_object(serializer)
        return cls(
            executor,
            spider,
            queue_key % {'spider': spider.name},
            serializer=serializer
        )

    async def len(self) -> int:
        return await self.container.get_message_count(self.key)

    async def push(self, request: aioscrapy.Request) -> None:
        data = self._encode_request(request)
        score = request.priority
        await self.container.publish(
            routing_key=self.key,
            body=data if isinstance(data, bytes) else data.encode(),
            priority=score
        )

    async def push_batch(self, requests: List[aioscrapy.Request]) -> None:
        # TODO: 实现rabbitmq的批量存储
        for request in requests:
            await self.push(request)

    async def pop(self, count: int = 1) -> Optional[aioscrapy.Request]:
        result = await self.container.get_message(self.key)
        if result:
            yield self._decode_request(result)

    async def clear(self) -> None:
        await self.container.clean_message_queue(self.key)


SpiderPriorityQueue = RabbitMqPriorityQueue



"""CloseSpider is an extension that forces spiders to be closed after certain
conditions are met.

See documentation in docs/topics/extensions.rst
"""
import asyncio
from typing import Optional
from collections import defaultdict

from aioscrapy import signals
from aioscrapy.exceptions import NotConfigured


class CloseSpider:

    def __init__(self, crawler):
        self.crawler = crawler

        self.close_on = {
            'timeout': crawler.settings.getfloat('CLOSESPIDER_TIMEOUT'),
            'itemcount': crawler.settings.getint('CLOSESPIDER_ITEMCOUNT'),
            'pagecount': crawler.settings.getint('CLOSESPIDER_PAGECOUNT'),
            'errorcount': crawler.settings.getint('CLOSESPIDER_ERRORCOUNT'),
        }

        if not any(self.close_on.values()):
            raise NotConfigured

        self.counter = defaultdict(int)
        self.task: Optional[asyncio.tasks.Task] = None

        if self.close_on.get('errorcount'):
            crawler.signals.connect(self.error_count, signal=signals.spider_error)
        if self.close_on.get('pagecount'):
            crawler.signals.connect(self.page_count, signal=signals.response_received)
        if self.close_on.get('timeout'):
            crawler.signals.connect(self.timeout_close, signal=signals.spider_opened)
        if self.close_on.get('itemcount'):
            crawler.signals.connect(self.item_scraped, signal=signals.item_scraped)
        crawler.signals.connect(self.spider_closed, signal=signals.spider_closed)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    async def error_count(self, failure, response, spider):
        self.counter['errorcount'] += 1
        if self.counter['errorcount'] == self.close_on['errorcount']:
            asyncio.create_task(self.crawler.engine.stop(reason='closespider_errorcount'))

    async def page_count(self, response, request, spider):
        self.counter['pagecount'] += 1
        if self.counter['pagecount'] == self.close_on['pagecount']:
            asyncio.create_task(self.crawler.engine.stop(reason='closespider_pagecount'))

    async def timeout_close(self, spider):
        async def close():
            await asyncio.sleep(self.close_on['timeout'])
            asyncio.create_task(self.crawler.engine.stop(reason='closespider_timeout'))

        self.task = asyncio.create_task(close())

    async def item_scraped(self, item, spider):
        self.counter['itemcount'] += 1
        if self.counter['itemcount'] == self.close_on['itemcount']:
            asyncio.create_task(self.crawler.engine.stop(reason='closespider_itemcount'))

    def spider_closed(self, spider):
        if self.task and not self.task.done():
            self.task.cancel()



# _*_ coding: utf-8 _*_

import asyncio
import logging
from typing import Optional, AsyncGenerator, Union, Callable
from asyncio import Queue
from asyncio.queues import QueueEmpty

import aioscrapy
from aioscrapy import Spider
from aioscrapy import signals
from aioscrapy.core.downloader import DownloaderTV
from aioscrapy.core.scheduler import BaseScheduler
from aioscrapy.core.scraper import Scraper
from aioscrapy.exceptions import DontCloseSpider
from aioscrapy.http import Response
from aioscrapy.http.request import Request
from aioscrapy.utils.log import logformatter_adapter
from aioscrapy.utils.misc import load_instance
from aioscrapy.utils.tools import call_helper

logger = logging.getLogger(__name__)


class Slot:

    def __init__(self, start_requests: Optional[AsyncGenerator]) -> None:
        self.inprogress: set[Request] = set()  # requests in progress
        self.start_requests = start_requests
        self.lock: bool = False

    def add_request(self, request: Request) -> None:
        self.inprogress.add(request)

    def remove_request(self, request: Request) -> None:
        self.inprogress.remove(request)


class ExecutionEngine(object):

    def __init__(self, crawler: "aioscrapy.Crawler") -> None:
        self.crawler = crawler
        self.settings = crawler.settings
        self.signals = crawler.signals
        self.logformatter = crawler.logformatter

        self.enqueue_cache_num = self.settings.getint("ENQUEUE_CACHE_NUM")
        self.enqueue_cache: Queue = Queue(self.enqueue_cache_num)
        self.slot: Optional[Slot] = None
        self.spider: Optional[Spider] = None
        self.downloader: Optional[DownloaderTV] = None
        self.scraper: Optional[Scraper] = None
        self.scheduler: Optional[BaseScheduler] = None

        self.running: bool = False
        self.unlock: bool = True
        self.finish: bool = False
        self.enqueue_unlock: bool = True

    async def start(
            self,
            spider: Spider,
            start_requests: Optional[AsyncGenerator] = None
    ) -> None:
        """Start the execution engine"""
        if self.running:
            raise RuntimeError("Engine already running")

        self.running = True
        await self.signals.send_catch_log_deferred(signal=signals.engine_started)
        await self.open(spider, start_requests)
        while not self.finish:
            self.running and await self._next_request()
            await asyncio.sleep(1)
            self.enqueue_cache_num != 1 and asyncio.create_task(self._crawl())
            self.running and await self._spider_idle(self.spider)

    async def stop(self, reason: str = 'shutdown') -> None:
        """Stop the execution engine gracefully"""
        if not self.running:
            raise RuntimeError("Engine not running")
        self.running = False

        while not self.is_idle():
            await asyncio.sleep(0.2)
            self.enqueue_cache_num != 1 and asyncio.create_task(self._crawl())
        await self.close_spider(self.spider, reason=reason)
        await self.signals.send_catch_log_deferred(signal=signals.engine_stopped)
        self.finish = True

    async def open(
            self,
            spider: Spider,
            start_requests: Optional[AsyncGenerator] = None
    ) -> None:
        logger.info("Spider opened", extra={'spider': spider})

        self.spider = spider
        await call_helper(self.crawler.stats.open_spider, spider)

        self.scheduler = await load_instance(self.settings['SCHEDULER'], crawler=self.crawler)
        self.downloader = await load_instance(self.settings['DOWNLOADER'], crawler=self.crawler)
        self.scraper = await call_helper(Scraper.from_crawler, self.crawler)

        start_requests = await call_helper(self.scraper.spidermw.process_start_requests, start_requests, spider)
        self.slot = Slot(start_requests)

        await self.signals.send_catch_log_deferred(signals.spider_opened, spider=spider)

    async def close(self) -> None:
        """Close the execution engine gracefully.

        If it has already been started, stop it. In all cases, close all spiders
        and the downloader.
        """
        if self.running:
            # Will also close spiders and downloader
            await self.stop()
        elif self.spider:
            # Will also close downloader
            await self.close_spider(self.spider, reason='shutdown')
        else:
            self.downloader.close()

    async def _next_request(self) -> None:
        assert self.slot is not None  # typing
        assert self.spider is not None  # typing

        while self.unlock and not self._needs_backout() and self.unlock:
            self.unlock = False
            try:
                async for request in self.scheduler.next_request(self.downloader.total_concurrency):
                    if request:
                        self.slot.add_request(request)
                        await self.downloader.fetch(request)
                break
            finally:
                self.unlock = True

        if self.slot.start_requests and not self._needs_backout() and not self.slot.lock:
            self.slot.lock = True
            try:
                request = await self.slot.start_requests.__anext__()
            except StopAsyncIteration:
                self.slot.start_requests = None
            except Exception as e:
                self.slot.start_requests = None
                logger.error('Error while obtaining start requests', exc_info=e, extra={'spider': self.spider})
            else:
                request and await self.crawl(request)
            finally:
                self.slot.lock = False

    def _needs_backout(self) -> bool:
        return (
                not self.running
                or self.downloader.needs_backout()
                or self.scraper.needs_backout()
        )

    async def handle_downloader_output(
            self, result: Union[Request, Response, BaseException, None], request: Request
    ) -> None:
        try:
            if result is None:
                return

            if not isinstance(result, (Request, Response, BaseException)):
                raise TypeError(
                    "Incorrect type: expected Request, Response or Failure, got %s: %r"
                    % (type(result), result)
                )

            if isinstance(result, Request):
                await self.crawl(result)
                return

            result.request = request
            if isinstance(result, Response):
                logkws = self.logformatter.crawled(request, result, self.spider)
                if logkws is not None:
                    logger.log(*logformatter_adapter(logkws), extra={'spider': self.spider})
                await self.signals.send_catch_log(signals.response_received,
                                                  response=result, request=request, spider=self.spider)
            await self.scraper.enqueue_scrape(result, request)

        finally:
            self.slot.remove_request(request)
            await self._next_request()

    def is_idle(self) -> bool:

        if self.downloader.active:
            # downloader has pending requests
            return False

        if self.slot.inprogress:
            # not all start requests are handled
            return False

        if not self.scraper.is_idle():
            # scraper is not idle
            return False

        return True

    async def crawl(self, request: Request) -> None:
        if self.enqueue_cache_num == 1:
            await self.scheduler.enqueue_request(request)
            asyncio.create_task(self._next_request())
        else:
            await self.enqueue_cache.put(request)

    async def _crawl(self) -> None:
        if not self.enqueue_unlock:
            return
        self.enqueue_unlock = False
        requests = []
        for _ in range(self.enqueue_cache.qsize()):
            try:
                request = self.enqueue_cache.get_nowait()
                requests.append(request)
            except QueueEmpty:
                break
        if requests:
            await call_helper(self.scheduler.enqueue_request_batch, requests)
            asyncio.create_task(self._next_request())
        self.enqueue_unlock = True

    async def close_spider(self, spider: Spider, reason: str = 'cancelled') -> None:
        """Close (cancel) spider and clear all its outstanding requests"""

        logger.info("Closing spider (%(reason)s)",
                    {'reason': reason},
                    extra={'spider': spider})

        async def close_handler(
                callback: Callable,
                *args,
                errmsg: str = '',
                **kwargs
        ) -> None:
            try:
                await call_helper(callback, *args, **kwargs)
            except (Exception, BaseException) as e:
                logger.error(
                    errmsg,
                    exc_info=e,
                    extra={'spider': spider}
                )

        await close_handler(self.downloader.close, errmsg='Downloader close failure')

        await close_handler(self.scraper.close, errmsg='Scraper close failure')

        await close_handler(self.scheduler.close, reason, errmsg='Scheduler close failure')

        await close_handler(self.signals.send_catch_log_deferred, signal=signals.spider_closed, spider=spider,
                            reason=reason, errmsg='Error while sending spider_close signal')

        await close_handler(self.crawler.stats.close_spider, spider, reason=reason, errmsg='Stats close failure')

        logger.info("Spider closed (%(reason)s)", {'reason': reason}, extra={'spider': spider})

        await close_handler(setattr, self, 'slot', None, errmsg='Error while unassigning slot')

        await close_handler(setattr, self, 'spider', None, errmsg='Error while unassigning spider')

    async def _spider_idle(self, spider: Spider) -> None:
        assert self.spider is not None
        res = await self.signals.send_catch_log(signals.spider_idle, spider=spider, dont_log=DontCloseSpider)
        if any(isinstance(x, DontCloseSpider) for _, x in res):
            return

        # method of 'has_pending_requests' has IO, so method of 'is_idle' execute twice
        if self.is_idle() \
                and self.slot.start_requests is None \
                and self.enqueue_unlock and self.enqueue_cache.empty() \
                and not await self.scheduler.has_pending_requests() \
                and self.is_idle():
            await self.stop(reason='finished')



import logging
import ssl

import httpx

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings

logger = logging.getLogger(__name__)


class HttpxDownloadHandler(BaseDownloadHandler):

    def __init__(self, settings):
        self.settings: Settings = settings
        self.httpx_client_session_args: dict = self.settings.get('HTTPX_CLIENT_SESSION_ARGS', {})
        self.verify_ssl: bool = self.settings.get("VERIFY_SSL")
        self.ssl_protocol = self.settings.get("SSL_PROTOCOL")  # ssl.PROTOCOL_TLSv1_2
        if self.settings.getbool("FIX_HTTPX_HEADER", True):
            # Fixed non-standard response's header 修复不标准的响应头
            import h11
            import re
            h11._readers.header_field_re = re.compile(b"(?P<field_name>.*?):[ \t](?P<field_value>.*?)")

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, _) -> HtmlResponse:
        kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'data': request.body or None
        }
        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        session_args = self.httpx_client_session_args.copy()
        session_args.update({
            'follow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect'),
            'max_redirects': self.settings.getint('REDIRECT_MAX_TIMES', 10),
        })
        ssl_ciphers = request.meta.get('TLS_CIPHERS')
        ssl_protocol = request.meta.get('ssl_protocol', self.ssl_protocol)
        if ssl_ciphers or ssl_protocol:
            if ssl_protocol:
                context = ssl.SSLContext(protocol=ssl_protocol)
            else:
                context = ssl.create_default_context()

            ssl_ciphers and context.set_ciphers(ssl_ciphers)
            session_args['verify'] = context

        proxy = request.meta.get("proxy")
        if proxy:
            session_args["proxies"] = proxy
            logger.debug(f"使用代理{proxy}抓取: {request.url}")

        async with httpx.AsyncClient(**session_args) as session:
            response = await session.request(request.method, request.url, **kwargs)
            content = response.read()

        return HtmlResponse(
            str(response.url),
            status=response.status_code,
            headers=response.headers,
            body=content,
            cookies=dict(response.cookies),
            encoding=response.encoding
        )

    async def close(self):
        pass



"""
This module implements the FormRequest class which is a more convenient class
(than Request) to generate Requests based on form data.

See documentation in docs/topics/request-response.rst
"""
from typing import List, Optional, Tuple, Union
from urllib.parse import urlencode

from aioscrapy.http.request import Request
from aioscrapy.utils.python import to_bytes, is_listlike

FormdataType = Optional[Union[dict, List[Tuple[str, str]]]]


class FormRequest(Request):
    valid_form_methods = ['GET', 'POST']

    def __init__(self, *args, formdata: FormdataType = None, **kwargs) -> None:
        if formdata and kwargs.get('method') is None:
            kwargs['method'] = 'POST'

        super().__init__(*args, **kwargs)

        if formdata:
            items = formdata.items() if isinstance(formdata, dict) else formdata
            form_query: str = _urlencode(items, self.encoding)
            if self.method == 'POST':
                self.headers.setdefault('Content-Type', 'application/x-www-form-urlencoded')
                self._set_body(form_query)
            else:
                self._set_url(self.url + ('&' if '?' in self.url else '?') + form_query)


def _urlencode(seq, enc):
    values = [(to_bytes(k, enc), to_bytes(v, enc))
              for k, vs in seq
              for v in (vs if is_listlike(vs) else [vs])]
    return urlencode(values, doseq=1)



import asyncio
import logging

import requests

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings

logger = logging.getLogger(__name__)


class RequestsDownloadHandler(BaseDownloadHandler):

    def __init__(self, settings):
        self.settings: Settings = settings
        self.verify_ssl: bool = self.settings.get("VERIFY_SSL")
        self.loop = asyncio.get_running_loop()

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, _) -> HtmlResponse:
        kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'verify': self.verify_ssl,
            'allow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect')
        }
        post_data = request.body or None
        if isinstance(post_data, dict):
            kwargs['json'] = post_data
        else:
            kwargs['data'] = post_data

        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        proxy = request.meta.get("proxy")
        if proxy:
            kwargs["proxies"] = {'http': proxy, 'https': proxy}
            logger.debug(f"use proxy {proxy}: {request.url}")

        response = await asyncio.to_thread(requests.request, request.method, request.url, **kwargs)
        return HtmlResponse(
            response.url,
            status=response.status_code,
            headers=response.headers,
            body=response.content,
            cookies=dict(response.cookies),
            encoding=response.encoding
        )

    async def close(self):
        pass



"""
Spider Middleware manager

See documentation in docs/topics/spider-middleware.rst
"""
from itertools import islice
from typing import AsyncIterable, Iterable, Callable, Union

from aioscrapy.exceptions import _InvalidOutput
from aioscrapy.utils.conf import build_component_list

from aioscrapy.http import Request, Response
from aioscrapy import Spider
from aioscrapy.settings import Settings
from aioscrapy.utils.tools import async_generator_wrapper
from aioscrapy.utils.tools import call_helper
from aioscrapy.middleware.absmanager import AbsMiddlewareManager


def _fname(f):
    return "{}.{}".format(
        f.__self__.__class__.__name__,
        f.__func__.__name__
    )


class SpiderMiddlewareManager(AbsMiddlewareManager):
    component_name = 'spider middleware'

    @classmethod
    def _get_mwlist_from_settings(cls, settings: Settings):
        return build_component_list(settings.getwithbase('SPIDER_MIDDLEWARES'))

    def _add_middleware(self, mw):
        super(SpiderMiddlewareManager, self)._add_middleware(mw)
        if hasattr(mw, 'process_spider_input'):
            self.methods['process_spider_input'].append(mw.process_spider_input)
        if hasattr(mw, 'process_start_requests'):
            self.methods['process_start_requests'].appendleft(mw.process_start_requests)
        process_spider_output = getattr(mw, 'process_spider_output', None)
        self.methods['process_spider_output'].appendleft(process_spider_output)
        process_spider_exception = getattr(mw, 'process_spider_exception', None)
        self.methods['process_spider_exception'].appendleft(process_spider_exception)

    async def scrape_response(self, scrape_func: Callable, response: Response, request: Request, spider: Spider):

        async def process_spider_input(response_) -> Union[AsyncIterable, Iterable]:
            for method in self.methods['process_spider_input']:
                try:
                    result = await call_helper(method, response=response_, spider=spider)
                    if result is not None:
                        raise _InvalidOutput(
                            f"Middleware {_fname(method)} must return None or raise an exception, got {type(result)}"
                        )
                except _InvalidOutput:
                    raise
                except BaseException as exception:
                    iterable_or_exception = await call_helper(scrape_func, exception, request)
                    if iterable_or_exception is exception:
                        raise iterable_or_exception
                    return iterable_or_exception
            return await call_helper(scrape_func, response_, request)

        async def _evaluate_iterable(result: Union[AsyncIterable, Iterable], exception_processor_index):
            try:
                # 将所有非AsyncGeneratorType变成AsyncGeneratorType对象
                async for r in await async_generator_wrapper(result):
                    yield r
            except BaseException as ex:
                exception_result = await process_spider_exception(ex, exception_processor_index)
                if isinstance(exception_result, BaseException):
                    raise exception_result
                async for r in _evaluate_iterable(exception_result, exception_processor_index):
                    yield r

        async def process_spider_exception(exception, start_index=0):
            # don't handle _InvalidOutput exception
            if isinstance(exception, _InvalidOutput):
                raise exception
            method_list = islice(self.methods['process_spider_exception'], start_index, None)
            for method_index, method in enumerate(method_list, start=start_index):
                if method is None:
                    continue
                result = await call_helper(method, response=response, exception=exception, spider=spider)
                if isinstance(result, AsyncIterable):
                    # stop exception handling by handing control over to the
                    # process_spider_output chain if an iterable has been returned
                    return await process_spider_output(result, method_index + 1)
                elif result is None:
                    continue
                else:
                    raise _InvalidOutput(
                            f"Middleware {_fname(method)}  must return None or an iterable, got {type(result)}"
                        )
            raise exception

        async def process_spider_output(result, start_index=0):
            # items in this iterable do not need to go through the process_spider_output
            # chain, they went through it already from the process_spider_exception method

            method_list = islice(self.methods['process_spider_output'], start_index, None)
            for method_index, method in enumerate(method_list, start=start_index):
                if method is None:
                    continue
                try:
                    # might fail directly if the output value is not a generator
                    result = await call_helper(method, response=response, result=result, spider=spider)
                except BaseException as ex:
                    exception_result = await process_spider_exception(ex, method_index + 1)
                    if isinstance(exception_result, BaseException):
                        raise
                    return exception_result
                if isinstance(result, AsyncIterable):
                    result = _evaluate_iterable(result, method_index + 1)
                else:
                    raise _InvalidOutput(f"Middleware {_fname(method)} must return an iterable, got {type(result)}")

            return result

        async def process_callback_output(result: Union[AsyncIterable, Iterable]):
            result: AsyncIterable = _evaluate_iterable(result, 0)
            return await process_spider_output(result)

        try:
            _iterable = await process_spider_input(response)
        except BaseException as exc:
            return await process_spider_exception(exc)
        else:
            return await process_callback_output(_iterable)

    async def process_start_requests(self, start_requests, spider):
        return await self._process_chain('process_start_requests', start_requests, spider)



"""
This module provides some useful functions for working with
scrapy.http.Response objects
"""
from typing import Iterable, Optional, Tuple, Union
from weakref import WeakKeyDictionary

from w3lib import html

import aioscrapy
from aioscrapy.http.response import Response

_baseurl_cache: "WeakKeyDictionary[Response, str]" = WeakKeyDictionary()


def get_base_url(response: "aioscrapy.http.response.TextResponse") -> str:
    """Return the base url of the given response, joined with the response url"""
    if response not in _baseurl_cache:
        text = response.text[0:4096]
        _baseurl_cache[response] = html.get_base_url(text, response.url, response.encoding)
    return _baseurl_cache[response]


_metaref_cache: "WeakKeyDictionary[Response, Union[Tuple[None, None], Tuple[float, str]]]" = WeakKeyDictionary()


def get_meta_refresh(
    response: "aioscrapy.http.response.TextResponse",
    ignore_tags: Optional[Iterable[str]] = ('script', 'noscript'),
) -> Union[Tuple[None, None], Tuple[float, str]]:
    """Parse the http-equiv refrsh parameter from the given response"""
    if response not in _metaref_cache:
        text = response.text[0:4096]
        _metaref_cache[response] = html.get_meta_refresh(
            text, response.url, response.encoding, ignore_tags=ignore_tags)
    return _metaref_cache[response]





import asyncio
import logging
from aioscrapy.db import db_manager

logger = logging.getLogger(__name__)


class SqlFormat:

    @staticmethod
    def ck_insert(table: str, fields: list, *args) -> str:
        fields = ','.join(fields)
        return f'''INSERT INTO {table} ({fields}) VALUES '''

    @staticmethod
    def mysql_insert(table: str, fields: list, *args) -> str:
        placeholder = ','.join(['%s'] * len(fields))
        fields = ','.join(fields)
        return f'''INSERT INTO {table} ({fields}) VALUES ({placeholder})'''

    @staticmethod
    def mysql_ignore_insert(table: str, fields: list, *args) -> str:
        placeholder = ','.join(['%s'] * len(fields))
        fields = ','.join(fields)
        return f'''INSERT IGNORE INTO {table} ({fields}) VALUES ({placeholder})'''

    @staticmethod
    def mysql_update_insert(table: str, fields: list, update_fields: list, *args) -> str:
        placeholder = ','.join(['%s'] * len(fields))
        if not update_fields:
            update_fields = fields
        update_fields = ','.join([f"{key} = VALUES({key})" for key in update_fields])
        fields = ','.join(fields)
        return f'INSERT INTO {table} ({fields}) VALUES ({placeholder}) ON DUPLICATE KEY UPDATE {update_fields}'

    def __call__(self, *args, db_type='mysql', insert_type='insert'):
        if getattr(self, f'{db_type}_{insert_type}'):
            func = getattr(self, f'{db_type}_{insert_type}')
            return func(*args)
        raise Exception(f"This write type is not supported： {db_type}_{insert_type}")


get_sql = SqlFormat()


class ItemCache(object):
    def __init__(self, db_type):
        self.db_type = db_type
        self.item_cache = {}
        self.fields_cache = {}
        self.table_cache = {}
        self.insert_sql_cache = {}
        self.db_alias_cache = {}

    def parse_item_to_cache(self, item: dict):
        table_name = item.pop('save_table_name')
        if table_name is None:
            raise Exception('please set save_table_name')
        insert_type = item.pop('save_insert_type', 'insert')
        update_fields = item.pop('save_update_fields', [])
        save_db_alias = item.pop('save_db_alias', ['default'])
        if isinstance(save_db_alias, str):
            save_db_alias = [save_db_alias]

        fields = list(item.keys())
        cache_key = ''.join(fields + update_fields + save_db_alias) + insert_type + table_name

        if self.fields_cache.get(cache_key) is None:
            self.db_alias_cache[cache_key] = save_db_alias
            self.table_cache[cache_key] = table_name
            self.fields_cache[cache_key] = fields
            self.item_cache[cache_key] = []
            self.insert_sql_cache[cache_key] = get_sql(
                table_name, fields, update_fields,
                db_type=self.db_type,
                insert_type=insert_type
            )

        self.item_cache[cache_key].append([item[field] for field in self.fields_cache[cache_key]])
        return cache_key, len(self.item_cache[cache_key])


class DBPipelineBase:
    def __init__(self, settings, db_type: str):
        self.cache_num = settings.getint('SAVE_CACHE_NUM', 500)
        self.save_cache_interval = settings.getint('SAVE_CACHE_INTERVAL', 10)
        self.db_type = db_type
        self.lock = asyncio.Lock()
        self.running: bool = True
        self.cache = ItemCache(db_type)

    async def open_spider(self, spider):
        asyncio.create_task(self.save_heartbeat())

    async def save_heartbeat(self):
        while self.running:
            await asyncio.sleep(self.save_cache_interval)
            asyncio.create_task(self.save_all())

    async def process_item(self, item, spider):
        await self.save_item(item)
        return item

    async def close_spider(self, spider):
        self.running = False
        await self.save_all()

    async def save_all(self):
        async with self.lock:
            for cache_key, items in self.cache.item_cache.items():
                items and await self._save(cache_key)

    async def save_item(self, item: dict):
        async with self.lock:
            cache_key, cache_count = self.cache.parse_item_to_cache(item)
            if cache_count >= self.cache_num:
                await self._save(cache_key)

    async def _save(self, cache_key):
        raise NotImplementedError


class MysqlPipeline(DBPipelineBase):

    @classmethod
    def from_settings(cls, settings):
        return cls(settings, 'mysql')

    async def _save(self, cache_key):
        table_name = self.cache.table_cache[cache_key]
        try:
            for alias in self.cache.db_alias_cache[cache_key]:
                async with db_manager.mysql.get(alias, ping=True) as (conn, cursor):
                    try:
                        num = await cursor.executemany(
                            self.cache.insert_sql_cache[cache_key], self.cache.item_cache[cache_key])
                        await conn.commit()
                        logger.info(f'table:{alias}->{table_name} sum:{len(self.cache.item_cache[cache_key])} ok:{num}')
                    except Exception as e:
                        await conn.rollback()
                        logger.exception(f'save data error, table:{alias}->{table_name}, err_msg:{e}')
        finally:
            self.cache.item_cache[cache_key] = []



"""
This module contains the default values for all settings used by Aioscrapy.

For more information about these settings you can read the settings
documentation in docs/topics/settings.rst

Aioscrapy developers, if you add a setting here remember to:

* add it in alphabetical order
* group similar settings without leaving blank lines
* add its documentation to the available settings documentation
  (docs/topics/settings.rst)

"""

import sys
from os.path import join, abspath, dirname

AUTOTHROTTLE_ENABLED = False
AUTOTHROTTLE_DEBUG = False
AUTOTHROTTLE_MAX_DELAY = 60.0
AUTOTHROTTLE_START_DELAY = 5.0
AUTOTHROTTLE_TARGET_CONCURRENCY = 1.0

BOT_NAME = 'aioscrapybot'

CLOSESPIDER_TIMEOUT = 0
CLOSESPIDER_PAGECOUNT = 0
CLOSESPIDER_ITEMCOUNT = 0
CLOSESPIDER_ERRORCOUNT = 0

COMMANDS_MODULE = ''

CONCURRENT_PARSER = 1

CONCURRENT_REQUESTS = 16
CONCURRENT_REQUESTS_PER_DOMAIN = 8
CONCURRENT_REQUESTS_PER_IP = 0


DEFAULT_REQUEST_HEADERS = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
    'Accept-Language': 'en',
}

DEPTH_LIMIT = 0
DEPTH_STATS_VERBOSE = False
DEPTH_PRIORITY = 0

DOWNLOAD_DELAY = 0

DOWNLOAD_HANDLERS = {}
DOWNLOAD_HANDLERS_BASE = {
    'http': 'aioscrapy.core.downloader.handlers.aiohttp.AioHttpDownloadHandler',
    'https': 'aioscrapy.core.downloader.handlers.aiohttp.AioHttpDownloadHandler',
}

DOWNLOAD_TIMEOUT = 180      # 3mins

DOWNLOADER = 'aioscrapy.core.downloader.Downloader'

DOWNLOADER_MIDDLEWARES = {}

DOWNLOADER_MIDDLEWARES_BASE = {
    # Engine side
    'aioscrapy.libs.downloader.downloadtimeout.DownloadTimeoutMiddleware': 350,
    'aioscrapy.libs.downloader.defaultheaders.DefaultHeadersMiddleware': 400,
    'aioscrapy.libs.downloader.useragent.UserAgentMiddleware': 500,
    'aioscrapy.libs.downloader.retry.RetryMiddleware': 550,
    'aioscrapy.libs.downloader.stats.DownloaderStats': 850,
    'aioscrapy.libs.downloader.ja3fingerprint.TLSCiphersMiddleware': 950,
    # Downloader side
}

DOWNLOADER_STATS = True

# DUPEFILTER_CLASS = 'aioscrapy.dupefilters.disk.RFPDupeFilter'
# DUPEFILTER_DEBUG = False

EDITOR = 'vi'
if sys.platform == 'win32':
    EDITOR = '%s -m idlelibs.idle'

EXTENSIONS = {}

EXTENSIONS_BASE = {
    'aioscrapy.libs.extensions.corestats.CoreStats': 0,
    'aioscrapy.libs.extensions.closespider.CloseSpider': 0,
    'aioscrapy.libs.extensions.logstats.LogStats': 0,
    'aioscrapy.libs.extensions.throttle.AutoThrottle': 0,
}


FILES_STORE_S3_ACL = 'private'
FILES_STORE_GCS_ACL = ''

HTTPPROXY_ENABLED = True
HTTPPROXY_AUTH_ENCODING = 'latin-1'

ITEM_PROCESSOR = 'aioscrapy.middleware.ItemPipelineManager'

ITEM_PIPELINES = {}
ITEM_PIPELINES_BASE = {}

LOG_ENABLED = True
LOG_ENCODING = 'utf-8'
LOG_MAX_BYTES = 50*1024*1024
LOG_BACKUP_COUNT = 10
LOG_FORMATTER = 'aioscrapy.logformatter.LogFormatter'
LOG_FORMAT = '%(asctime)s %(process)d [%(name)s] %(levelname)s: %(message)s'
LOG_DATEFORMAT = '%Y-%m-%d %H:%M:%S'
LOG_STDOUT = False
LOG_LEVEL = 'DEBUG'
LOG_FILE = None
LOG_SHORT_NAMES = False

SCHEDULER_DEBUG = False

LOGSTATS_INTERVAL = 60.0


METAREFRESH_ENABLED = True
METAREFRESH_IGNORE_TAGS = []
METAREFRESH_MAXDELAY = 100

NEWSPIDER_MODULE = ''

RANDOMIZE_DOWNLOAD_DELAY = True

REDIRECT_ENABLED = True
REDIRECT_MAX_TIMES = 20

REFERER_ENABLED = True
REFERRER_POLICY = 'aioscrapy.libs.spider.referer.DefaultReferrerPolicy'

RETRY_ENABLED = True
RETRY_TIMES = 2  # initial response + 2 retries = 3 requests
RETRY_HTTP_CODES = [500, 502, 503, 504, 522, 524, 408, 429]
RETRY_PRIORITY_ADJUST = -1

SCHEDULER = 'aioscrapy.core.scheduler.Scheduler'
SCHEDULER_QUEUE_CLASS = 'aioscrapy.queue.memory.SpiderPriorityQueue'
SCHEDULER_SERIALIZER = 'aioscrapy.serializer.JsonSerializer'

SCRAPER_SLOT_MAX_ACTIVE_SIZE = 5000000

SPIDER_LOADER_CLASS = 'aioscrapy.spiderloader.SpiderLoader'
SPIDER_LOADER_WARN_ONLY = False

SPIDER_MIDDLEWARES = {}
SPIDER_MIDDLEWARES_BASE = {
    'aioscrapy.libs.spider.httperror.HttpErrorMiddleware': 50,
    'aioscrapy.libs.spider.offsite.OffsiteMiddleware': 500,
    'aioscrapy.libs.spider.referer.RefererMiddleware': 700,
    'aioscrapy.libs.spider.urllength.UrlLengthMiddleware': 800,
    'aioscrapy.libs.spider.depth.DepthMiddleware': 900,
}

SPIDER_MODULES = []

STATS_CLASS = 'aioscrapy.statscollectors.MemoryStatsCollector'
STATS_DUMP = True

TEMPLATES_DIR = abspath(join(dirname(__file__), '..', 'templates'))

URLLENGTH_LIMIT = 2083

CLOSE_SPIDER_ON_IDLE = False



import logging

from motor.motor_asyncio import AsyncIOMotorClient

import aioscrapy
from aioscrapy.db.absmanager import AbsDBPoolManager

logger = logging.getLogger(__name__)


class MongoExecutor:
    def __init__(self, alias: str, pool_manager: "AioMongoManager"):
        self.alias = alias
        self.pool_manager = pool_manager

    async def insert(self, table_name, values, db_name=None):
        client, db_name_default = self.pool_manager.get_pool(self.alias)
        db_name = db_name or db_name_default
        return await client[f'{db_name}'][f'{table_name}'].insert_many(values)

    def __getattr__(self, table_name: str):
        client, db_name_default = self.pool_manager.get_pool(self.alias)
        return client[f'{db_name_default}'][f'{table_name}']


class AioMongoManager(AbsDBPoolManager):
    _clients = {}

    async def create(self, alias: str, params: dict):
        if alias in self._clients:
            return self._clients[alias]

        params = params.copy()
        db_name = params.pop('db')
        params.setdefault('connecttimeoutms', 30)
        client = AsyncIOMotorClient(**params)
        return self._clients.setdefault(alias, (client, db_name))

    def get_pool(self, alias: str):
        return self._clients.get(alias)

    def executor(self, alias: str) -> MongoExecutor:
        """Get RedisExecutor"""
        return MongoExecutor(alias, self)

    async def close(self, alias: str):
        """Close mongo pool named `alias`"""
        client, *_ = self._clients.pop(alias, None)
        if client:
            client.close()

    async def close_all(self):
        """Close all clients of mongo"""
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_dict(self, db_args: dict):
        """Create mongo with dict"""
        for alias, args in db_args.items():
            await self.create(alias, args)

    async def from_settings(self, settings: aioscrapy.Settings):
        """Create mongo with settings"""
        for alias, args in settings.getdict('MONGO_ARGS').items():
            await self.create(alias, args)


mongo_manager = AioMongoManager()

if __name__ == '__main__':
    import asyncio


    async def test():
        await mongo_manager.create('default', {
            'host': 'mongodb://root:root@192.168.234.128:27017',
            'db': 'test',
        })
        mongo = mongo_manager.executor('default')
        result = await mongo.insert('user', [{'name': 'zhang', 'age': 18}, {'name': 'li', 'age': 20}])
        # print('inserted %d docs' % (len(result.inserted_ids),))

        document = await mongo.user.find_one({'img_url': {'$gt': 19}})
        print(document)
        await mongo_manager.close_all()


    asyncio.get_event_loop().run_until_complete(test())



from collections.abc import Mapping


class Headers(dict):
    """Case insensitive http headers dictionary"""

    __slots__ = ()

    def __init__(self, seq=None):
        super().__init__()
        if seq:
            self.update(seq)

    def __getitem__(self, key):
        return dict.__getitem__(self, self.normkey(key))

    def __setitem__(self, key, value):
        dict.__setitem__(self, self.normkey(key), self.normvalue(value))

    def __delitem__(self, key):
        dict.__delitem__(self, self.normkey(key))

    def __contains__(self, key):
        return dict.__contains__(self, self.normkey(key))

    has_key = __contains__

    def __copy__(self):
        return self.__class__(self)

    copy = __copy__

    def normkey(self, key):
        """Method to normalize dictionary key access"""
        return key.title()

    def normvalue(self, value):
        """Method to normalize values prior to be setted"""
        return value

    def get(self, key, def_val=None):
        return dict.get(self, self.normkey(key), self.normvalue(def_val))

    def setdefault(self, key, def_val=None):
        return dict.setdefault(self, self.normkey(key), self.normvalue(def_val))

    def update(self, seq):
        seq = seq.items() if isinstance(seq, Mapping) else seq
        iseq = ((self.normkey(k), self.normvalue(v)) for k, v in seq)
        super().update(iseq)

    @classmethod
    def fromkeys(cls, keys, value=None):
        return cls((k, value) for k in keys)

    def pop(self, key, *args):
        return dict.pop(self, self.normkey(key), *args)



import traceback
import warnings
from collections import defaultdict

from zope.interface import Interface, implementer

from aioscrapy.utils.misc import walk_modules
from aioscrapy.utils.spider import iter_spider_classes


class ISpiderLoader(Interface):

    def from_settings(settings):
        """Return an instance of the class for the given settings"""

    def load(spider_name):
        """Return the Spider class for the given spider name. If the spider
        name is not found, it must raise a KeyError."""

    def list():
        """Return a list with the names of all spiders available in the
        project"""

    def find_by_request(request):
        """Return the list of spiders names that can handle the given request"""


@implementer(ISpiderLoader)
class SpiderLoader:
    """
    SpiderLoader is a class which locates and loads spiders
    in a aioscrapy project.
    """

    def __init__(self, settings):
        self.spider_modules = settings.getlist('SPIDER_MODULES')
        self.warn_only = settings.getbool('SPIDER_LOADER_WARN_ONLY')
        self._spiders = {}
        self._found = defaultdict(list)
        self._load_all_spiders()

    def _check_name_duplicates(self):
        dupes = []
        for name, locations in self._found.items():
            dupes.extend([
                f"  {cls} named {name!r} (in {mod})"
                for mod, cls in locations
                if len(locations) > 1
            ])

        if dupes:
            dupes_string = "\n\n".join(dupes)
            warnings.warn(
                "There are several spiders with the same name:\n\n"
                f"{dupes_string}\n\n  This can cause unexpected behavior.",
                category=UserWarning,
            )

    def _load_spiders(self, module):
        for spcls in iter_spider_classes(module):
            self._found[spcls.name].append((module.__name__, spcls.__name__))
            self._spiders[spcls.name] = spcls

    def _load_all_spiders(self):
        for name in self.spider_modules:
            try:
                for module in walk_modules(name):
                    self._load_spiders(module)
            except ImportError:
                if self.warn_only:
                    warnings.warn(
                        f"\n{traceback.format_exc()}Could not load spiders "
                        f"from module '{name}'. "
                        "See above traceback for details.",
                        category=RuntimeWarning,
                    )
                else:
                    raise
        self._check_name_duplicates()

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    def load(self, spider_name):
        """
        Return the Spider class for the given spider name. If the spider
        name is not found, raise a KeyError.
        """
        try:
            return self._spiders[spider_name]
        except KeyError:
            raise KeyError(f"Spider not found: {spider_name}")

    def find_by_request(self, request):
        """
        Return the list of spider names that can handle the given request.
        """
        return [
            name for name, cls in self._spiders.items()
            if cls.handles_request(request)
        ]

    def list(self):
        """
        Return a list with the names of all spiders available in the project.
        """
        return list(self._spiders.keys())



import logging
from abc import ABCMeta, abstractmethod

from aioscrapy.utils.python import global_object_name

logger = logging.getLogger('aioscrapy.proxy')


class AbsProxy(metaclass=ABCMeta):
    def __init__(self, settings):
        self.use_proxy = settings.getbool('USE_PROXY', False)
        self.max_count = settings.getint('PROXY_MAX_COUNT', 16)
        self.min_count = settings.getint('PROXY_MIN_COUNT', 1)
        self.allow_status_code = settings.get('PROXY_ALLOW_STATUS_CODE', [404])
        self.cache = []

    async def add_proxy(self, request):
        """add proxy for request"""
        if self.use_proxy and request.use_proxy:
            request.meta['proxy'] = await self.get()
        else:
            request.meta.pop('proxy', None)
        return request

    def remove(self, proxy, reason=None):
        if callable(reason):
            reason = reason()
        if isinstance(reason, Exception):
            reason = global_object_name(reason.__class__)

        if proxy in self.cache:
            logger.info(f"remove proxy: {proxy}, reason: {reason}")
            self.cache.remove(proxy)

    def check(self, request, response=None, exception=None):
        if not self.use_proxy:
            return

        if response and response.status >= 400 and response.status not in self.allow_status_code:
            self.remove(request.meta['proxy'], f"Don't allow response status code:{response.status}")

        if exception and isinstance(exception, BaseException):
            self.remove(request.meta['proxy'], exception)

    @classmethod
    @abstractmethod
    async def from_crawler(cls, crawler) -> "AbsProxy":
        """get proxy instance from spider"""

    @abstractmethod
    async def get(self) -> str:
        """get proxy"""



# _*_ coding: utf-8 _*_
import asyncio
from types import CoroutineType, GeneratorType, AsyncGeneratorType


async def call_helper(f, *args, **kwargs):
    if asyncio.iscoroutinefunction(f):
        return await f(*args, **kwargs)
    return f(*args, **kwargs)


async def async_generator_wrapper(wrapped):
    """将传入的对象变成AsyncGeneratorType类型"""
    if isinstance(wrapped, AsyncGeneratorType):
        return wrapped

    elif isinstance(wrapped, CoroutineType):
        async def anonymous(c):
            yield await c
        return anonymous(wrapped)

    elif isinstance(wrapped, GeneratorType):
        async def anonymous(c):
            for r in c:
                yield r
        return anonymous(wrapped)

    else:
        async def anonymous(c):
            yield c
        return anonymous(wrapped)


def singleton(cls):
    _instance = {}

    def _singleton(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]

    return _singleton


def exec_js_func(js_file_path, func_name, func_params=None, cwd_path=None, cmd_path='node'):
    """
    调用node执行js文件
    :param js_file_path: js文件路径
    :param func_name: 要调用的js文件内的函数名称
    :param func_params: func的参数
    :param cwd_path: node_modules文件所在路径，如果不指定将使用全局的node_modules
    :param cmd_path: node的位置， 例如：r'D:\\path\to\node.exe'
    :return: func_name函数的执行结果
    """
    import execjs

    if func_params is None:
        func_params = []
    name = "MyNode"
    execjs.register(name, execjs._external_runtime.ExternalRuntime(
        name="Node.js (V8)",
        command=[cmd_path],
        encoding='UTF-8',
        runner_source=execjs._runner_sources.Node
    ))
    with open(js_file_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    js = ''.join(lines)
    js_context = execjs.get(name).compile(js, cwd=cwd_path)
    return js_context.call(func_name, *func_params)
